@@ -59,6 +59,7 @@
 #include "src/common/xsignal.h"
 #include "src/common/xstring.h"
 
+#include "src/interfaces/conn.h"
 #include "src/interfaces/cred.h"
 
 #define STDIO_MAX_FREE_BUF 1024
@@ -79,7 +80,8 @@ static struct io_buf *_alloc_io_buf(void);
 static void	_init_stdio_eio_objs(slurm_step_io_fds_t fds,
 				     client_io_t *cio);
 static void	_handle_io_init_msg(int fd, client_io_t *cio);
-static int _read_io_init_msg(int fd, client_io_t *cio, slurm_addr_t *host);
+static int _read_io_init_msg(int fd, void *conn, client_io_t *cio,
+			     slurm_addr_t *host);
 static int      _wid(int n);
 static bool     _incoming_buf_free(client_io_t *cio);
 static bool     _outgoing_buf_free(client_io_t *cio);
@@ -88,7 +90,7 @@ static bool     _outgoing_buf_free(client_io_t *cio);
  * Listening socket declarations
  **********************************************************************/
 static bool _listening_socket_readable(eio_obj_t *obj);
-static int _listening_socket_read(eio_obj_t *obj, List objs);
+static int _listening_socket_read(eio_obj_t *obj, list_t *objs);
 
 struct io_operations listening_socket_ops = {
 	.readable = &_listening_socket_readable,
@@ -99,9 +101,9 @@ struct io_operations listening_socket_ops = {
  * IO server socket declarations
  **********************************************************************/
 static bool _server_readable(eio_obj_t *obj);
-static int _server_read(eio_obj_t *obj, List objs);
+static int _server_read(eio_obj_t *obj, list_t *objs);
 static bool _server_writable(eio_obj_t *obj);
-static int _server_write(eio_obj_t *obj, List objs);
+static int _server_write(eio_obj_t *obj, list_t *objs);
 
 struct io_operations server_ops = {
 	.readable = &_server_readable,
@@ -124,7 +126,7 @@ struct server_io_info {
 	int remote_stderr_objs; /* active eio_obj_t's on the remote node */
 
 	/* outgoing variables */
-	List msg_queue;
+	list_t *msg_queue;
 	struct io_buf *out_msg;
 	int32_t out_remaining;
 	bool out_eof;
@@ -134,7 +136,7 @@ struct server_io_info {
  * File write declarations
  **********************************************************************/
 static bool _file_writable(eio_obj_t *obj);
-static int _file_write(eio_obj_t *obj, List objs);
+static int _file_write(eio_obj_t *obj, list_t *objs);
 
 struct io_operations file_write_ops = {
 	.writable = &_file_writable,
@@ -145,7 +147,7 @@ struct file_write_info {
 	client_io_t *cio;
 
 	/* outgoing variables */
-	List msg_queue;
+	list_t *msg_queue;
 	struct io_buf *out_msg;
 	int32_t out_remaining;
 	/* If taskid is (uint32_t)-1, output from all tasks is accepted,
@@ -159,7 +161,7 @@ struct file_write_info {
  * File read declarations
  **********************************************************************/
 static bool _file_readable(eio_obj_t *obj);
-static int _file_read(eio_obj_t *obj, List objs);
+static int _file_read(eio_obj_t *obj, list_t *objs);
 
 struct io_operations file_read_ops = {
 	.readable = &_file_readable,
@@ -196,8 +198,7 @@ _listening_socket_readable(eio_obj_t *obj)
 	return true;
 }
 
-static int
-_listening_socket_read(eio_obj_t *obj, List objs)
+static int _listening_socket_read(eio_obj_t *obj, list_t *objs)
 {
 	client_io_t *cio = (client_io_t *)obj->arg;
 
@@ -218,9 +219,9 @@ _set_listensocks_nonblocking(client_io_t *cio)
 /**********************************************************************
  * IO server socket functions
  **********************************************************************/
-static eio_obj_t *
-_create_server_eio_obj(int fd, client_io_t *cio, int nodeid,
-		       int stdout_objs, int stderr_objs)
+static eio_obj_t *_create_server_eio_obj(int fd, void *conn, client_io_t *cio,
+					 int nodeid, int stdout_objs,
+					 int stderr_objs)
 {
 	eio_obj_t *eio = NULL;
 	struct server_io_info *info = xmalloc(sizeof(*info));
@@ -240,6 +241,7 @@ _create_server_eio_obj(int fd, client_io_t *cio, int nodeid,
 
 	net_set_keep_alive(fd);
 	eio = eio_obj_create(fd, &server_ops, (void *)info);
+	eio->conn = conn;
 
 	return eio;
 }
@@ -284,8 +286,7 @@ _server_readable(eio_obj_t *obj)
 	return false;
 }
 
-static int
-_server_read(eio_obj_t *obj, List objs)
+static int _server_read(eio_obj_t *obj, list_t *objs)
 {
 	struct server_io_info *s = (struct server_io_info *) obj->arg;
 	void *buf;
@@ -300,7 +301,7 @@ _server_read(eio_obj_t *obj, List objs)
 			return SLURM_ERROR;
 		}
 
-		n = io_hdr_read_fd(obj->fd, &s->header);
+		n = io_hdr_read_fd(obj->fd, obj->conn, &s->header);
 		if (n <= 0) { /* got eof or error on socket read */
 			if (n < 0) {	/* Error */
 				if (obj->shutdown) {
@@ -369,7 +370,12 @@ _server_read(eio_obj_t *obj, List objs)
 	if (s->header.length != 0) {
 		buf = s->in_msg->data + (s->in_msg->length - s->in_remaining);
 	again:
-		if ((n = read(obj->fd, buf, s->in_remaining)) < 0) {
+		if (obj->conn) {
+			n = conn_g_recv(obj->conn, buf, s->in_remaining);
+		} else {
+			n = read(obj->fd, buf, s->in_remaining);
+		}
+		if (n < 0) {
 			if (errno == EINTR)
 				goto again;
 			if ((errno == EAGAIN) || (errno == EWOULDBLOCK))
@@ -463,8 +469,7 @@ _server_writable(eio_obj_t *obj)
 	return false;
 }
 
-static int
-_server_write(eio_obj_t *obj, List objs)
+static int _server_write(eio_obj_t *obj, list_t *objs)
 {
 	struct server_io_info *s = (struct server_io_info *) obj->arg;
 	void *buf;
@@ -494,7 +499,12 @@ _server_write(eio_obj_t *obj, List objs)
 	 */
 	buf = s->out_msg->data + (s->out_msg->length - s->out_remaining);
 again:
-	if ((n = write(obj->fd, buf, s->out_remaining)) < 0) {
+	if (obj->conn) {
+		n = conn_g_send(obj->conn, buf, s->out_remaining);
+	} else {
+		n = write(obj->fd, buf, s->out_remaining);
+	}
+	if (n < 0) {
 		if (errno == EINTR) {
 			goto again;
 		} else if ((errno == EAGAIN) || (errno == EWOULDBLOCK)) {
@@ -569,7 +579,7 @@ static bool _file_writable(eio_obj_t *obj)
 	return false;
 }
 
-static int _file_write(eio_obj_t *obj, List objs)
+static int _file_write(eio_obj_t *obj, list_t *objs)
 {
 	struct file_write_info *info = (struct file_write_info *) obj->arg;
 	void *ptr;
@@ -690,7 +700,7 @@ static bool _file_readable(eio_obj_t *obj)
 	return false;
 }
 
-static int _file_read(eio_obj_t *obj, List objs)
+static int _file_read(eio_obj_t *obj, list_t *objs)
 {
 	struct file_read_info *info = (struct file_read_info *) obj->arg;
 	struct io_buf *msg;
@@ -710,10 +720,10 @@ static int _file_read(eio_obj_t *obj, List objs)
 	}
 	slurm_mutex_unlock(&info->cio->ioservers_lock);
 
-	ptr = msg->data + io_hdr_packed_size();
+	ptr = msg->data + IO_HDR_PACKET_BYTES;
 
 again:
-	if ((len = read(obj->fd, ptr, MAX_MSG_LEN)) < 0) {
+	if ((len = read(obj->fd, ptr, SLURM_IO_MAX_MSG_LEN)) < 0) {
 		if (errno == EINTR)
 			goto again;
 		if ((errno == EAGAIN) || (errno == EWOULDBLOCK)) {
@@ -740,9 +750,9 @@ again:
 	 */
 	header = info->header;
 	header.length = len;
-	packbuf = create_buf(msg->data, io_hdr_packed_size());
+	packbuf = create_buf(msg->data, IO_HDR_PACKET_BYTES);
 	io_hdr_pack(&header, packbuf);
-	msg->length = io_hdr_packed_size() + header.length;
+	msg->length = IO_HDR_PACKET_BYTES + header.length;
 	msg->ref_count = 0; /* make certain it is initialized */
 	/* free the packbuf structure, but not the memory to which it points */
 	packbuf->head = NULL;
@@ -790,7 +800,7 @@ again:
 
 
 /**********************************************************************
- * General fuctions
+ * General functions
  **********************************************************************/
 
 static void *
@@ -835,11 +845,12 @@ _create_listensock_eio(int fd, client_io_t *cio)
 	return eio;
 }
 
-static int _read_io_init_msg(int fd, client_io_t *cio, slurm_addr_t *host)
+static int _read_io_init_msg(int fd, void *conn, client_io_t *cio,
+			     slurm_addr_t *host)
 {
 	io_init_msg_t msg = { 0 };
 
-	if (io_init_msg_read_from_fd(fd, &msg) != SLURM_SUCCESS)
+	if (io_init_msg_read_from_fd(fd, conn, &msg) != SLURM_SUCCESS)
 		goto fail;
 
 	if (io_init_msg_validate(&msg, cio->io_key) < 0) {
@@ -861,9 +872,10 @@ static int _read_io_init_msg(int fd, client_io_t *cio, slurm_addr_t *host)
 		error("IO: Hey, you told me node %d was down!", msg.nodeid);
 	}
 
-	cio->ioserver[msg.nodeid] = _create_server_eio_obj(fd, cio, msg.nodeid,
-							   msg.stdout_objs,
-							   msg.stderr_objs);
+	cio->ioserver[msg.nodeid] =
+		_create_server_eio_obj(fd, conn, cio, msg.nodeid,
+				       msg.stdout_objs, msg.stderr_objs);
+
 	slurm_mutex_lock(&cio->ioservers_lock);
 	bit_set(cio->ioservers_ready_bits, msg.nodeid);
 	cio->ioservers_ready = bit_set_count(cio->ioservers_ready_bits);
@@ -882,6 +894,7 @@ static int _read_io_init_msg(int fd, client_io_t *cio, slurm_addr_t *host)
 	return SLURM_SUCCESS;
 
     fail:
+	conn_g_destroy(conn, false);
 	xfree(msg.io_key);
 	if (fd > STDERR_FILENO)
 		close(fd);
@@ -912,6 +925,7 @@ _handle_io_init_msg(int fd, client_io_t *cio)
 
 	for (j = 0; j < 15; j++) {
 		int sd;
+		void *conn = NULL;
 		slurm_addr_t addr;
 
 		/*
@@ -920,9 +934,7 @@ _handle_io_init_msg(int fd, client_io_t *cio)
 		if (!_is_fd_ready(fd))
 			return;
 
-		while ((sd = slurm_accept_msg_conn(fd, &addr)) < 0) {
-			if (errno == EINTR)
-				continue;
+		while (!(conn = slurm_accept_msg_conn(fd, &addr))) {
 			if (errno == EAGAIN)	/* No more connections */
 				return;
 			if ((errno == ECONNABORTED) ||
@@ -933,6 +945,7 @@ _handle_io_init_msg(int fd, client_io_t *cio)
 			return;
 		}
 
+		sd = conn_g_get_fd(conn);
 		debug3("Accepted IO connection: ip=%pA sd=%d", &addr, sd);
 
 		/*
@@ -947,7 +960,7 @@ _handle_io_init_msg(int fd, client_io_t *cio)
 		/*
 		 * Read IO header and update cio structure appropriately
 		 */
-		if (_read_io_init_msg(sd, cio, &addr) < 0)
+		if (_read_io_init_msg(sd, conn, cio, &addr) < 0)
 			continue;
 
 		fd_set_nonblocking(sd);
@@ -973,7 +986,7 @@ _alloc_io_buf(void)
 	buf->length = 0;
 	/* The following "+ 1" is just temporary so I can stick a \0 at
 	   the end and do a printf of the data pointer */
-	buf->data = xmalloc(MAX_MSG_LEN + io_hdr_packed_size() + 1);
+	buf->data = xmalloc(SLURM_IO_MAX_MSG_LEN + IO_HDR_PACKET_BYTES + 1);
 
 	return buf;
 }
@@ -1309,11 +1322,11 @@ int client_io_handler_send_test_message(client_io_t *cio, int node_id,
 	if (_incoming_buf_free(cio)) {
 		msg = list_dequeue(cio->free_incoming);
 
-		msg->length = io_hdr_packed_size();
+		msg->length = IO_HDR_PACKET_BYTES;
 		msg->ref_count = 1;
 		msg->header = header;
 
-		packbuf = create_buf(msg->data, io_hdr_packed_size());
+		packbuf = create_buf(msg->data, IO_HDR_PACKET_BYTES);
 		io_hdr_pack(&header, packbuf);
 		/* free the packbuf, but not the memory to which it points */
 		packbuf->head = NULL;
